#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import socket
import websocket
from websocket import create_connection
from multiprocessing import Process, Queue, JoinableQueue
import asyncio
import threading
import configparser
import os
import sys
import time
import functools
import struct
import ctypes
import inspect
import numpy as np
import pandas as pd

from npsdkobjs import *
from Stockdrv import *

import logging
LOG_FORMAT = ('%(levelname) -10s %(asctime)s %(name) -30s %(funcName) '
              '-35s %(lineno) -5d: %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(level = logging.INFO)
#logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
formatter = logging.Formatter(LOG_FORMAT)

handler = logging.FileHandler("npdatapoolthread.log")
handler.setLevel(logging.INFO)
handler.setFormatter(formatter)

console = logging.StreamHandler()
console.setLevel(logging.INFO)
console.setFormatter(formatter)
logger.addHandler(handler)
#logger.addHandler(console)

doc="""
	    #历史K线标准周期：分时、分笔、1分K线、5分钟K线、日线。15、30、60分钟由5分钟组成；周、月、季、年K线由日线生成。

	    #初始化：市场代码表、除权、财务、实时行情等，第一次连接，自动申请一次，市场代码有变动，主服务器会主动推送。
        #init = "类型=初始化&编码=unicode"  # unicode 比 utf-8 快很多
	    #self.Send(init)

	    #复权：0 无复权或不填写(默认)；-1 向前复权：历史方向(通常做法)；1 向后复权：未来方向

	    #当天数据：0 不包括当天数据；
	    #格式：0 结构体(默认)、1 Json
        #setPqram  = "类型=K线参数&当天数据=1&复权=0&格式=0&版本=OEM_Ver2020"  #按约定格式传回K线数据
        #self.Send(setPqram)

        #split    = "代码=SH600000&类型=除权"
        #finance  = "代码=SH600000&类型=财务"
        #splitFin = "代码=SH600000&类型=除权财务"
        #f10       = "代码=SH600000&类型=F10资料"
        #report   = "代码=SH600000&类型=实时行情"
        
        #print("正在申请%s \r\n" % (split))
        #self.send_command_over_ws(split)
        #self.send_command_over_ws(finance)
        #self.send_command_over_ws(splitFin)
        #self.send_command_over_ws(f10)
        #self.send_command_over_ws(report)
        
        #day      = "代码=SH600000&类型=日线&数量=0&开始=0"  #申请日线，数量：K线根数，0 使用面板设置值；开始：从最新往历史方向第几根
        day      = "代码=SH000001&类型=日线&数量=1000&开始=1000"  #申请日线，数量：K线根数，0 使用面板设置值；开始：从最新往历史方向第几根
        #week     = "代码=SH000001&类型=周线&数量=5000"         #申请周线(未支持)
        #month    = "代码=SH600000&类型=月线&数量=5000"         #申请月线(未支持)
        #quarter  = "代码=SH600000&类型=季线&数量=5000"         #申请季线(未支持)
        #year     = "代码=SH600000&类型=年线&数量=5000"         #申请年线(未支持)
        min1     = "代码=SH600000&类型=1分钟线&数量=3000"      #申请1分钟线
        min5     = "代码=SH600000&类型=5分钟线&数量=3000"      #申请5分钟线
        #min15    = "代码=SH600000&类型=15分钟线&数量=3000"     #申请15分钟线(未支持)
        #min30    = "代码=SH600000&类型=30分钟线&数量=3000"     #申请30分钟线(未支持)
        #min60    = "代码=SH600000&类型=60分钟线&数量=3000"     #申请60分钟线(未支持)
        #trace    = "代码=SH600000&类型=分笔&数量=0"            #申请分笔(每3秒)
        #tick     = "代码=SH600000&类型=分时&数量=3000"         #分时(每分钟)(未支持)
        
        #print("正在申请%s \r\n" % (day))
        #self.send_command_over_ws(day)
        #self.send_command_over_ws(week)
        #self.send_command_over_ws(month)
        #self.send_command_over_ws(quarter)
        #self.send_command_over_ws(year)
        #self.send_command_over_ws(min1)
        #self.send_command_over_ws(min5)
        #self.send_command_over_ws(min15)
        #self.send_command_over_ws(min30)
	    #self.send_command_over_ws(min60)
        #self.send_command_over_ws(trace)
        #self.send_command_over_ws(tick)
        
        #wprintf("正在申请%s \r\n", askDay)
        #self.send_command_over_ws(askDay)
        
        #提示：申请某个板块，可以自行建立一个函数，连续申请，后台自动向服务器申请数据。
        
        #close  = "类型=关闭接口"
        #hide   = "类型=隐藏接口"
        #show   = "类型=显示接口"
        #market = "类型=市场代码表"
        
        #self.send_command_over_ws(close)
        #self.send_command_over_ws(hide)
        #self.send_command_over_ws(show)
        #self.send_command_over_ws(market)
"""

class NpDataPoolThread(threading.Thread): #(Process)
    
    internal_lock = threading.Lock()

    def __init__(self,  command_queue, response_queue, exePpath = ".\\网际风\\"):

        threading.Thread.__init__(self)
        #super().__init__()

        self._command_queue = command_queue     # command queue for requesting data
        self._response_queue = response_queue   # response queue for sending data
        
        self._ws_message_queue = JoinableQueue()    # queue for data received from websocket  
        self._ws_command_queue = JoinableQueue()    # queue for command sent to websocket
        logger.info('get and create internal queues...')

        self.processqueuethread_stopevent = threading.Event()
        self.sendwscommandthread_stopevent = threading.Event()

        self._stopped_flag = False 

        #os.environ["CUDA_DEVICES_ORDER"]   = "PCI_BUS_IS"
        #os.environ["CUDA_VISIBLE_DEVICES"] = '0'  #表示调用0块GPU
        self.m_ws   = None
        self.m_stop = False
        self.m_url  = "ws://127.0.0.1:39398/"
        self.m_logined = NPSDKAPI_WS_DISCONNECTED
        #websocket.enableTrace(True)

        file = exePpath + "用户\\配置文件.ini"

        config = configparser.ConfigParser() # 创建配置文件对象
        config.read(file, encoding='utf-16') # 读取文件
        ip   = config.get('第三方调用', '连接地址01')
        port = config.get('第三方调用', '端口01')
        self.m_url = 'ws://' + ip +':' + port


    """    
    向_ws_message_queue中压入从ws收到的数据
    返回：bool
    """
    def put_ws_message_queue(self, msg):  
        try:
            #with self.internal_lock:                
                if self._stopped_flag: 
                    return False 

                if not self._ws_message_queue.full():
                    self._ws_message_queue.put(msg) #put_nowait()
                    return True
                else:
                    return False
        except Exception as e:
            logger.warning('put_wsmessage_queue() exception: %r'% e)
            return False

    #停止线程函数
    def _async_raise(self, tid, exctype):
        """raises the exception, performs cleanup if needed"""
        tid = ctypes.c_long(tid)
        if not inspect.isclass(exctype):
            exctype = type(exctype)
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
        print("res:%s"% res)
        if res == 0:
            raise ValueError("invalid thread id")
        elif res != 1:
            # """if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect"""
            ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
            raise SystemError("PyThreadState_SetAsyncExc failed")

    #是否K线格式
    def is_klines_type(self, type)-> bool:   
        arr = ["1分钟线", "5分钟线", "15分钟线", "30分钟线", "60分钟线", "日线", "周线", "月线", "季线", "年线", "多日线"]
        #x = arr.count(type)
        return type in arr

    #发送基于websocket的数据请求命令
    def send_command_over_ws(self, ask):
        if self.m_logined == NPSDKAPI_WS_DISCONNECTED:   #网络断开
            print("数据接口连接已断开")
            return
        try:
            ask = "股票数据?" + ask + "&版本=20210310"
            #logger.info('wscmd: '+ask)
            self.m_ws.send(ask)
        except Exception as e:
            error = "数据接口连接已断开"
            self.on_error(error)

            logger.warning('send_command_over_ws() exception: %r'% e)
            
    def on_message(self, message): #实时全推数据处理量大，建议拷贝数据后，扔到队列去 其它线程去处理
        """ 
        把原始数据压入_ws_message_queue队列，留给子线程后续处理
        """
        # 纪录历史数据包，以备数据包回测用
        #with open('backtest0602-1.dat','ab') as fo:
        #    fo.write(message)
        #fo.close()

        #self._df_wsincoming_rawdata.loc[time.time()] = message
        #compression_opts = dict(method='zip', archive_name='out.csv')  
        #self._df_wsincoming_rawdata.to_csv('out.zip', index=False, compression=compression_opts) 
        
        #print('-------------------on_message type: %s '%type(message))
        self.put_ws_message_queue(message)

    def on_error(self, error):
        print(error)

    def on_close(self):
        print("关闭数据接口连接")
        self.m_logined = NPSDKAPI_WS_DISCONNECTED

    def on_open(self):
        print("数据接口连接成功")
        self.m_logined = NPSDKAPI_WS_CONNECTED

        init = "类型=初始化&编码=unicode"  # unicode 比 utf-8 快很多
        self.send_command_over_ws(init)

        #当天数据：0 不包括当天数据；
	    #格式：0 结构体(默认)、1 Json
        #setPqram  = "类型=K线参数&当天数据=1&复权=0&格式=0&版本=OEM_Ver2020"  #按约定格式传回K线数据
        setPqram = "类型=K线参数&当天数据=1&复权=0&格式=0"  # 按约定格式传回K线数据
        self.send_command_over_ws(setPqram)

    def Connect(self):
        try:
            self.m_ws = create_connection(self.m_url, sockopt = ((socket.IPPROTO_TCP, socket.TCP_NODELAY, 1),))
        except Exception as e:
            error = "数据接口连接失败"
            self.on_error(error)
            self.on_close()
            return
        try:
            self.on_open()
            while True:
                if self.m_stop: 
                    break         
                rcv = self.m_ws.recv()
                self.on_message(rcv)
        except Exception as e:
            error = "数据接口连接已断开"
            self.on_error(error)
            
            logger.warning('websocket exception:%r'% e)
        finally:
            self.m_stop = True
            self.on_close()
            self.m_ws.close()
            return

    def start_me(self):

        # 启动队列处理子线程 
        self.processqueuethread_stopevent.clear()
        self.processqueuethread = threading.Thread(target=self.process_queue_thread, args=(self.processqueuethread_stopevent,))
        self.processqueuethread.start()
        
        self.sendwscommandthread_stopevent.clear()
        self.sendwscommandthread = threading.Thread(target=self.send_wscommand_thread, args=(self.sendwscommandthread_stopevent,))
        self.sendwscommandthread.start()  
 
        """ 主线程体 """
        while True:

            self.Connect()

            if self.m_stop: 
                break

            time.sleep(0.1)
        #print("websocket close and exit from the loop")
    
    
    def stop_me(self):
        self._stopped_flag = True #停止使用Queue发送消息

        self.m_logined = NPSDKAPI_WS_DISCONNECTED
        self.m_stop = True
        self.on_close()
        self.m_ws.close() #关闭websocket

        self.processqueuethread_stopevent.set() # 停止该线程: 设置Event为True,使之从while loop中退出
        self.processqueuethread.join()
        self.sendwscommandthread_stopevent.set() # 停止该线程: 设置Event为True,使之从while loop中退出
        self.sendwscommandthread.join()

        # 如果使用了Queue, Python会派生daemon性质的QueueFeederThread, 需要释放Queue
        self._ws_message_queue.join()
        self._ws_message_queue.close()
        self._ws_command_queue.join()
        self._ws_command_queue.close()
        #self._async_raise(self.sendwscommandthread.ident, SystemExit)
        logger.info("当前线程数量: %d  \r\n线程是：%s" % (len(threading.enumerate()), str(threading.enumerate())))

    """
    主线程入口函数：调用npdatapoolthread.start()即进入run()
    """
    def run(self):
        self.start_me()  



    # ----------------------------------------------------------------------
    """""""""""""""""""""""""""""""""""""""""""""
    子线程（队列处理线程）: 功能：
    1. 接收并分析来自websocket的业务数据, 
       更新内存数据池中的业务数据字典
    .....
    2. 接收并分析来自npsdkapi的命令
    """""""""""""""""""""""""""""""""""""""""""""
    def process_queue_thread(self, stop_event):

        """ added by xxg in 2021.5.5
        本地业务数据内存截面
        1. 市场代码表 
        2. 实时行情 
        3. ...
        4. ...
        """
        glb_marketinfo_dict = {} # 市场表 证券代码表 数据字典
        glb_marketinfo_dict.clear()

        glb_stk_report_dict = {} # 证券实时行情 数据字典
                                        # {
                                        #   'SH600000':OEM_REPORT,
                                        #   'SH600007':OEM_REPORT,
                                        #   ......
                                        #   'SH600100':OEM_REPORT,
                                        #   'SH600101':OEM_REPORT,
                                        # }
        glb_stk_report_dict.clear()

        #["1分钟线", "5分钟线", "15分钟线", "30分钟线", "60分钟线", "日线", "周线", "月线", "季线", "年线", "多日线"]
        glb_stk_klines_dict = {} # 证券K线数据 数据字典 
                                        # {                      [RCV_KLINE,RCV_KLINE,RCV_KLINE,]
                                        #   'SH600000':{'1分钟线':原始数据字节流,    
                                        #               '5分钟线':原始数据字节流, 
                                        #               ......
                                        #              },
                                        #   ......
                                        #   'SH600007':{'1分钟线':原始数据字节流,
                                        #               '日线':   原始数据字节流, 
                                        #               ......
                                        #              },
                                        # }
        glb_stk_klines_dict.clear()

        glb_wscommands_set = set()       # websocket commands set


        """    
        从_ws_message_queue中取出wsmessage
        返回：元组(bool, message)
        """
        def get_wsmessage_queue():   
            wsmessage = ''
            
            try:
                #with self.internal_lock:    
                    if self._stopped_flag: 
                        return False, wsmessage 
                
                    if not self._ws_message_queue.empty():
                        wsmessage = self._ws_message_queue.get()  #get_nowait() 
                        #self._ws_message_queue.task_done()
                        return True, wsmessage  
                    else:
                        return False, wsmessage  
            except Exception as e:
                logger.warning('get_wsmessage_queue() exception: %r'% e)
                return False, wsmessage 


        """""""""""""""""""""""""""""""""""""""""""""    
        检查来自websocket的数据
        """""""""""""""""""""""""""""""""""""""""""""
        def check_queue_from_npdatapool():

            bret, wsmessage = get_wsmessage_queue()
            if bret: #成功取得一条message   
                
                if len(wsmessage) == 0 or len(wsmessage) < sizeof(OEM_DATA_HEAD) :
                    # 接收到的消息长度为0 或者 小于OEM_DATA_HEAD结构体长度，则为错误消息，立即返回
                    print('错误消息: 接收到的消息 长度为0或者小于OEM_DATA_HEAD结构体长度')
                    logger.error('错误消息: 接收到的消息 长度为0或者小于OEM_DATA_HEAD结构体长度')
                    return 

                """
                message = HEAD + BODY
                HEAD: (100字节)
                """
                wshead = OEM_DATA_HEAD()
                wshead.decode(wsmessage)  #把字节流 转成 ctypes结构体
                
                wsheadinfo = "ws data head info:\r\n[oemVer:%s] x [type:%s] x [label:%s] x [name:%s] x [flag:%s] x [askId:%s] x [len:%s] x [count:%s]" \
                % (wshead.oemVer, wshead.type, wshead.label, wshead.name, wshead.flag, wshead.askId, wshead.len, wshead.count)
                #print(wsheadinfo)
                logger.info(wsheadinfo)

                """
                message = HEAD + BODY
                BODY: (可变长度)
                """
                wsbody = wsmessage
                wsbody = wsmessage[sizeof(OEM_DATA_HEAD) : ] # 取出 wsbody

                msgtype   = wshead.type            #消息类型
                #headsize   = sizeof(OEM_DATA_HEAD)  #100 字节

                """
                # 这里默认为websocket server 一次性发送的字节流都是完整的业务数据流，不存在没发完的残留数据
                # 需要大消息体测试 (比如一次性收 几千根Kline数据)
                """ 

                if msgtype == '代码表':
                    pass
                
                
                #-"""实时行情数据处理"""-#
                elif msgtype == '实时数据':  
                    if len(wsbody) == 0 or len(wsbody) != wshead.count * sizeof(OEM_REPORT):
                        # 接收到的实时行情消息体长度为0 或者 实际长度不等于count*sizeof(OEM_REPORT)，则为错误消息，立即返回
                        print('错误消息: 接收到的实时行情消息体长度为0 或者 实际长度不等于count*sizeof(OEM_REPORT)')
                        logger.error('错误消息: 接收到的实时行情消息体长度为0 或者 实际长度不等于count*sizeof(OEM_REPORT)')
                        return 

                    for i in range(wshead.count):
                        reportbody = OEM_REPORT()
                        reportbody.decode(wsbody)               #字节流 转成 ctypes结构体
                        wsbody = wsbody[sizeof(OEM_REPORT) : ]  #删除一个结构体
                        #print(reportbody.label, reportbody.name)
                        #print('wsbody len%s'% len(wsbody))

                        """
                        待加功能：在代码表里检查 reportbody.label 是否为有效证券代码
                        """
                        
                         #以证券代码为key值 更新实时行情字典 
                        glb_stk_report_dict[reportbody.label] = reportbody   
                        #print('更新实时行情数据字典: %s len of dict：%s'% (reportbody.label, len(glb_stk_report_dict)))

                        """
                        if reportbody.label == 'SH600000' :
                            tm = time.localtime(reportbody.time)
                            tmStr = time.strftime('%Y-%m-%d %H:%M:%S', tm)
                            print("接收到实时行情: %s(%s) : time:%s close : %0.2f" % (reportbody.label, reportbody.name, tmStr, reportbody.close))
                        """
                    #print('---------------len(report dict):%s'%len(glb_stk_report_dict))
                
                #-"""K线数据处理"""-#
                elif self.is_klines_type(msgtype):

                    #logger.info(wsbody) 
                    if len(wsbody) == 0 or len(wsbody) != wshead.count * sizeof(RCV_KLINE):
                        # 接收到的实时行情消息体长度为0 或者 实际长度不等于count*sizeof(RCV_KLINE)，则为错误消息，立即返回
                        print('错误消息: 接收到的K线数据消息体长度为0 或者 实际长度不等于count*sizeof(RCV_KLINE)')
                        logger.error('错误消息: 接收到的K线数据消息体长度为0 或者 实际长度不等于count*sizeof(RCV_KLINE)')
                        return

                    symbol = wshead.label       #证券代码
                    klines_type = wshead.type   #Klines类型字符串

                    """
                    待加功能：在代码表里检查 wshead.label 是否为有效证券代码
                    """

                    #以证券代码为key值 更新K线数据一级字典。以K线类型为key值 更新K线数据二级字典。
                    if  symbol in glb_stk_klines_dict.keys():
                        glb_stk_klines_dict[symbol][klines_type] = wsbody 
                    else: # 该证券代码第一次传送K线数据
                        klines_dict = {}
                        klines_dict[klines_type] = wsbody
                        glb_stk_klines_dict[symbol] = klines_dict
                    
                    """
                    if wshead.label == 'SH600000' :
                        tm = time.localtime()
                        tmStr = time.strftime('%Y-%m-%d %H:%M:%S', tm)
                        print("接收到K线数据: %s(%s) : time:%s " % (wshead.label, wshead.name, tmStr))
                    """
                
                elif msgtype == '分时' or msgtype == '分笔' :
                    pass
                
                elif msgtype == '除权':
                    pass

                elif msgtype == '财务' :
                    pass

                elif msgtype == 'F10资料' :                 
                    pass


        """    
        从command_queue中取出命令
        返回：元组(bool, NpApiDataObj)
        """
        def get_command_queue():   
            npapidataobj = NpApiDataObj()
            
            try:
                #with self.internal_lock:                   
                    if self._stopped_flag: 
                        return False, npapidataobj 

                    if not self._command_queue.empty():
                        npapidataobj = self._command_queue.get()  #get_nowait() 
                        #self._command_queue.task_done()
                        return True, npapidataobj  
                    else:
                        return False, npapidataobj  
            except Exception as e:
                logger.warning('get_command_queue() exception: %r'% e)
                return False, npapidataobj 

        """    
        向response_queue中压入命令数据响应
        返回：bool
        """
        def put_response_queue(npapidataobj):  
            try:
                #with self.internal_lock:                      
                    if self._stopped_flag: 
                        return False 
             
                    if not self._response_queue.full():
                        self._response_queue.put(npapidataobj) #put_nowait()
                        return True
                    else:
                        return False
            except Exception as e:
                logger.warning('put_response_queue() exception: %r'% e)
                return False

        """    
        向ws_command_queue中压入wscommand命令
        返回：bool
        """
        def put_wscommands_queue(_wscommands):  
            try:
                #with self.internal_lock:                      
                    if self._stopped_flag: 
                        return False 
             
                    if not self._ws_command_queue.full():
                        self._ws_command_queue.put(_wscommands) #put_nowait()
                        return True
                    else:
                        return False
            except Exception as e:
                logger.warning('put_wscommands_queue() exception: %r'% e)
                return False
        
        """""""""""""""""""""""""""""""""""""""""""""    
        检查是否有NpSdkApi队列任务
        """""""""""""""""""""""""""""""""""""""""""""
        def check_queue_from_npsdkapi():

            bret, npapidataobj = get_command_queue()
            #print('datapool: cmd_queue(size:%s:%s) msg_queue(size:%s)'%(self._command_queue.qsize(), bret, self._ws_message_queue.qsize()))

            if bret: #成功取得一条命令
                app_id = npapidataobj.app_id

                if app_id == NPSDKAPI_APPID_NPSDKAPI: #该消息为策略程序请求命令
                    #print("got command from command_queue...app_id: %s, message_id: %s, request_body: %s"%\
                    #    (npapidataobj.app_id, npapidataobj.message_id, npapidataobj.request_body))
                    
                    if npapidataobj.message_id == NPSDKAPI_MESSAGEID_GETQUOTE_REQ: 
                        """
                        获取指定代码的实时行情
                        """
                        bsuc, npapiobj = get_quote_remote(npapidataobj) 
                        
                        """
                        返回指定代码的实时行情数据
                        """
                        if bsuc:
                            put_response_queue(npapiobj)

                   
                    elif npapidataobj.message_id == NPSDKAPI_MESSAGEID_GETKLINES_REQ: 
                        """
                        获取指定代码的K线数据
                        """
                        bsuc, npapiobj = get_kline_serial_remote(npapidataobj) 
                        
                        """
                        返回指定代码的K线数据
                        """
                        if bsuc:
                            put_response_queue(npapiobj)

                        
                    elif npapidataobj.message_id == NPSDKAPI_MESSAGEID_WSCOMMAND_REQ: #向websocket server发送命令请求数据
                        
                        # 监控队列长度，如果队列长度太大表示消息堆积
                        if self._ws_message_queue.qsize() > NPSDKAPI_QUEUESZIE_WARNING:
                            logger.warning('(wsmsg_queue: %s) (wscmd_queue: %s) (response_queue: %s) (command_queue: %s)' \
                                %(self._ws_message_queue.qsize(), self._ws_command_queue.qsize(), self._response_queue.qsize(), self._command_queue.qsize()))
                        #print('wsmsg_queue:%s wscmd_queue:%s response_queue:%s command_queue:%s'%(self._ws_message_queue.qsize(), self._ws_command_queue.qsize(), self._response_queue.qsize(), self._command_queue.qsize()))
                    
                        logger.info('send ws command: %s' % npapidataobj.request_body)
                        logger.info(npapidataobj.request_body)
                        self.send_command_over_ws(npapidataobj.request_body)

                        glb_wscommands_set.add(npapidataobj.request_body) # 汇集各路websocket commands命令用set去重
                        #print(glb_wscommands_set)
                        #put_wscommands_queue(glb_wscommands_set)

                    else: #其它命令
                        pass

                else: # 没有收到来自 NPSDKAPI_APPID_NPSDKAPI 的命令
                    pass


        """
        get_quote_remote函数体
        返回：元组（bool, NpApiDataObj）
        """
        def get_quote_remote(npapiobjsrc):
            """ 
            在数据池里查找实时行情业务数据，
            """         
            npapiobjdes = NpApiDataObj()
            npapiobjdes.app_id = NPSDKAPI_APPID_NPSDKDATA
            npapiobjdes.message_id = NPSDKAPI_MESSAGEID_GETQUOTE_RES
            npapiobjdes.correlation_id = npapiobjsrc.correlation_id
            npapiobjdes.reply_to = npapiobjsrc.reply_to
            npapiobjdes.response_type = ''
            npapiobjdes.response_body = ''

            try:                                      
                codekey = npapiobjsrc.request_body#(npapiobjsrc.request_body).decode()
            except Exception as e:
                logprtstr ='get_quote_remote() exception with codekey: %r'% e
                logger.warning(logprtstr)
                npapiobjdes.response_body = 'error: ' + logprtstr
                npapiobjdes.response_type = NPSDKAPI_ERROR
                return False, npapiobjdes

            # 快速查找实时行情字典
            if  codekey in glb_stk_report_dict: 
                reportbody = glb_stk_report_dict[codekey]
                #print("found codekey:%s label: %s in 实时行情字典"% (codekey, reportbody.label))
                
                if reportbody:
                    #print("在实时行情字典中找到 %s 开盘价(%s) 最新价(%s)" % (reportbody.label, reportbody.open, reportbody.close))
                    #logger.info("在实时行情字典中找到codekey")
                    
                    npapiobjdes.response_body = reportbody.encode() # 把ctypes结构体转成字节流
                    npapiobjdes.response_type = NPSDKAPI_SUCCESS

                    return True, npapiobjdes 
                else:
                    logprtstr = "在实时行情字典中找到codekey：%s，但是对应的数据体为空"% codekey
                    logger.warning(logprtstr)
                    npapiobjdes.response_body = 'error: ' + logprtstr 
                    npapiobjdes.response_type = NPSDKAPI_ERROR
                    return False, npapiobjdes
            
            else:
                logprtstr = "在实时行情字典中没找到codekey:%s "% codekey
                logger.warning(logprtstr)
                npapiobjdes.response_body = 'error: ' +  logprtstr
                npapiobjdes.response_type = NPSDKAPI_ERROR  
                return False, npapiobjdes

        """
        get_kline_serial_remote函数体
        返回：元组（bool, NpApiDataObj）
        """
        def get_kline_serial_remote(npapiobjsrc):
            """ 
            在数据池里查找K线业务数据，
            """         
            npapiobjdes = NpApiDataObj()
            npapiobjdes.app_id = NPSDKAPI_APPID_NPSDKDATA
            npapiobjdes.message_id = NPSDKAPI_MESSAGEID_GETKLINES_RES
            npapiobjdes.correlation_id = npapiobjsrc.correlation_id
            npapiobjdes.reply_to = npapiobjsrc.reply_to
            npapiobjdes.response_type = ''
            npapiobjdes.response_body = ''

            try:                               
                # 分析参数  symbol + '?' + klines_type_str  # 格式: 格式: 'SH600000?1分钟线'  
                requestbody = npapiobjsrc.request_body.split('?')
                codekey = requestbody[0] #指定代码
                klines_type = requestbody[1] #K线类型

            except Exception as e:
                logprtstr = 'get_kline_serial_remote() exception with requestbody: %r'% e
                logger.warning(logprtstr)
                npapiobjdes.response_body = 'error: ' + logprtstr
                npapiobjdes.response_type = NPSDKAPI_ERROR
                return False, npapiobjdes

            # 快速查找K线字典: 在一级字典找到指定代码且同时在二级字典找到K线类别
            
            if  codekey in glb_stk_klines_dict.keys():
                if klines_type in glb_stk_klines_dict[codekey].keys():
                    kline_rawdata = glb_stk_klines_dict[codekey][klines_type]
                    
                    if kline_rawdata:
                        npapiobjdes.response_body = kline_rawdata
                        npapiobjdes.response_type = NPSDKAPI_SUCCESS
                        return True, npapiobjdes 
                    else:
                        logprtstr = "在K线数据字典中找到(codekey %s:klinestype %s), 但是对应的数据体为空"% (codekey, klines_type)
                        logger.warning(logprtstr)
                        npapiobjdes.response_body = 'error: ' + logprtstr
                        npapiobjdes.response_type = NPSDKAPI_ERROR
                        return False, npapiobjdes
                else:
                    logprtstr = "在K线数据字典中找到codekey %s 但没找到klinestype %s"% (codekey, klines_type)
                    logger.warning(logprtstr)
                    npapiobjdes.response_body = 'error: ' + logprtstr
                    npapiobjdes.response_type = NPSDKAPI_ERROR
                    return False, npapiobjdes
            else:
                logprtstr = "在K线数据字典中没找到codekey:%s "% codekey
                logger.warning(logprtstr)
                npapiobjdes.response_body = 'error: ' + logprtstr
                npapiobjdes.response_type = NPSDKAPI_ERROR  
                return False, npapiobjdes

        """
        子线程（队列处理线程）线程体
        """
        while not stop_event.is_set():
            # check queue if has ws message from npdata and update data in memory dict
            check_queue_from_npdatapool()

            # check queue from npsdkapi if has data request
            check_queue_from_npsdkapi()

            """
            由于websocket的数据消息可能会很多, check_queue_from_npdatapool()函数会占有大量时间片
            为了尽快处理来自npsdkapi发来的_command_queue任务，sleep值不能太大
            """
            time.sleep(0.0001)  
        #print('process_queue_thread exit from the loop')




    # ----------------------------------------------------------------------
    """
    子线程（基于websocket的数据请求命令线程）: 功能：
    1. 根据业务类型设置定时器任务发送数据请求命令 （避免频繁发送命令）
    2. 发送基于websocket的数据请求命令
    """
    def send_wscommand_thread(self, stop_event): 

        """    
        从_ws_command_queue中取出wscommand
        返回：元组(bool, command)
        """
        def get_wscommands_queue()->set:   
            _wscommands = set()
            
            try:
                #with self.internal_lock:    
                    if self._stopped_flag: 
                        return False, _wscommands 
                
                    if not self._ws_command_queue.empty():
                        _wscommands = self._ws_command_queue.get()  #get_nowait() 
                        #self._ws_command_queue.task_done()
                        return True, _wscommands  
                    else:
                        return False, _wscommands  
            except Exception as e:
                logger.warning('get_wscommands_queue() exception: %r'% e)
                return False, _wscommands 

        #子线程（队列处理线程）线程体
        while not stop_event.is_set():
            
            #print('wscommand_queue:%s '%self._ws_command_queue.qsize())
            bret, wscommands = get_wscommands_queue() #把wscommand加入集合达到去重的目的
            if bret: #成功取得一个wscommand命令集合
                #print(wscommands)

                #for wsc in wscommands:
                #    logger.info('send ws command: %s' % wsc)
                #    logger.info(wsc)
                #    self.send_command_over_ws(wsc)
                #    time.sleep(0.2)
                pass

            time.sleep(3)
    
    """
    def OnAsk(self):
        if self.m_logined != NPSDKAPI_WS_CONNECTED:
            return
    """

 
    






        



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
"""
NpSdkApi接口的PYTHON封装
"""
__author__ = 'xxg'

from multiprocessing import Process, Queue, JoinableQueue
import copy
import functools
import json
import os
import re
import ssl
import sys
import time
import uuid
from datetime import datetime
from typing import Union, List, Any, Optional
import numpy as np
import pandas as pd
from ctypes import *

from npsdkobjs import *
from Stockdrv import *
from npdatapoolthread import NpDataPoolThread

#from objs import Quote , Kline, Tick, Account, Position, Order, Trade
#from log import _get_log_format, _get_log_name, _clear_logs

from __version__ import __version__

import logging
LOG_FORMAT = ('%(levelname) -10s %(asctime)s %(name) -30s %(funcName) '
              '-35s %(lineno) -5d: %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(level = logging.INFO)
#logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
formatter = logging.Formatter(LOG_FORMAT)

handler = logging.FileHandler("npsdkapi.log")
handler.setLevel(logging.INFO)
handler.setFormatter(formatter)

console = logging.StreamHandler()
console.setLevel(logging.INFO)
console.setFormatter(formatter)
logger.addHandler(handler)
#logger.addHandler(console)

class NpSdkApi(object):
    """
    NpSdkApi接口及数据管理类
    """

    def __init__(self, auth: Optional[str] = None, debug: bool = False) -> None:
                 
        doc="""
        创建NpSdkApi接口实例
        """

        # NpSDK Rpc Client
        #self._npsdkrpcclient = NpSdkRpcClient()
        #logger.info('create NpSdkRpcClient...')

        # 内部关键数据

        # 记录历史命令请求 以及 最后一次的数据响应 
        self._npsdkapi_cmd_request_response = {
            NPSDKAPI_MESSAGEID_GETQUOTE_REQ: {}, #      'get_quote':{   证券代码:[命令参数,实时行情quote,读取数据成功与否,策略程序实时行情quote]}  二层字典 
                                                 #      'get_quote':{   'SH600000':[cmdpara,quote,bsuc,quote1st], 
                                                 #                      'SH600007':[cmdpara,quote,bsuc,quote1st],
                                                 #                      ..........
                                                 #                  }
                                                 #    
            NPSDKAPI_MESSAGEID_GETKLINES_REQ: {}, #     'get_klines':{  证券代码:{K线类别：[cmdpara,K线数据df,读取数据成功与否,策略程序K线数据df,]}} 三层字典 
                                                #       "get_klines":{  "SH600000": {'1分钟线':[cmdpara,klinesdf,bsuc,klinesdf1st],'5分钟线':[cmdpara,klinesdf,bsuc,klinesdf1st],...,},
                                                #                       "SH600001": {'1分钟线':[cmdpara,klinesdf,bsuc,klinesdf1st],'5分钟线':[cmdpara,klinesdf,bsuc,klinesdf1st],...,},
                                                #                       "SH600002": {'1分钟线':[cmdpara,klinesdf,bsuc,klinesdf1st],'5分钟线':[cmdpara,klinesdf,bsuc,klinesdf1st],...,},
                                                #                       ......
                                                #                    }
        }   
        # 28个字段定义见 Stockdrv.py文件里的 class OEM_REPORT(Structure):
        # 可以通过语句注释功能来选择字段判断实时行情数据是否更新    
        self._fieldsOfQuoteToCompared = [      #实时行情 400 字节，带*表示网际风内部数据，请跳过
                        'label',            #代码
                        'name',             #名称
                        'time',             #成交时间UTC时间，可以百度搜索相关转换)
                        'foot',             #最后处理分钟数
                        'openDate',         #市场开盘日期零点
                        'openTime',         #市场开盘时间
                        'closeDate',        #市场日线日期零点
                        'open',             #今日开盘
                        'high',             #今日最高
                        'low',              #今日最低
                        'close',            #最新价格
                        'volume',           #总成交量
                        'amount',           #总成交金额
                        #'inVol',            #外盘*
                        'pricesell',        #申卖价1,2,3,4,5
                        'volsell',           #申卖量1,2,3,4,5      
                        #'vsellCha',         #申卖量变化*
                        'pricebuy',         #申买价1,2,3,4,5      
                        'volbuy',            #申买量1,2,3,4,5       
                        #'vbuyCha',           #申买量变化*
                        'jingJia',          #集合竞价
                        'avPrice',          #期货的结算价（平均价）
                        #'isBuy',            #0 下跌；1 上涨或平盘*
                        'nowv',             #现量
                        'nowa',             #分笔成交额期货仓差)
                        #'change',           #换手率*
                        #'weiBi',            #委比*
                        #'liangBi',          #量比*
                        #'temp',       
                    ] # fields

        self._typesOfKline = ["1分钟线", "5分钟线", "15分钟线", "30分钟线", "60分钟线", "日线", "周线", "月线", "季线", "年线", "多日线"]
        
        # 字段定义见 Stockdrv.py文件里的 class RCV_KLINE(Structure):
        self._fieldsOfKline = [      
                    'time',                 #时间 UTC时间戳 单位：s秒
                    'open',                 #开盘
                    'high',                 #最高
                    'low',                  #最低
                    'close',                #现价
                    'volume',               #成交量
                    'amount',               #成交额
                    'temp',                 #预留
                    ] # fields

        self._stopped_flag = False

        self._debug_npsdk = debug

        self.run_datapool_thread() #启动datapoolthread

    def run_datapool_thread(self):
        self._command_queue = JoinableQueue()     
        self._response_queue = JoinableQueue()    
        logger.info('create global command queue & response queue')

        # 创建数据仓库进程
        self._npdatapoolthread = NpDataPoolThread(self._command_queue, self._response_queue)
        self._npdatapoolthread.start() 
        
        logger.info('create & start npdatapool thread')


    def close(self) -> None:
        """
        关闭NpSdkApi接口实例并释放相应资源
        """
        self._stopped_flag = True #停止使用Queue发送消息
        
        # 暴力退出
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)

        self._npdatapoolthread.stop_me() 

        self._command_queue.join()
        self._command_queue.close()
        self._response_queue.join()
        self._response_queue.close()
        
 
    """    
    向command_queue中压入请求命令
    返回：bool
    """
    def put_command_queue(self, npapidataobj):  
        try:
            #with self.internal_lock:                
                if self._stopped_flag: 
                    return False 
                if not self._command_queue.full():
                    self._command_queue.put(npapidataobj) #put_nowait()
                    #print('put command into command_queue ')
                    return True
                else:
                    return False
        except Exception as e:
            logger.warning('put_command_queue() exception: %r'% e)
            return False

    """    
    从response_queue中取出命令数据响应
    返回：元组(bool, NpApiDataObj)
    """
    def get_response_queue(self) -> NpApiDataObj:   
        npapidataobj = NpApiDataObj()
        
        try:
            #with self.internal_lock:                
                if self._stopped_flag: 
                    return False, npapidataobj             
                if not self._response_queue.empty():
                    npapidataobj = self._response_queue.get()  #get_nowait() 
                    #print('got response from response_queue ')
                    #self._response_queue.task_done()
                    return True, npapidataobj  
                else:
                    return False, npapidataobj  
        except Exception as e:
            logger.warning('get_response_queue() exception: %r'% e)
            return False, npapidataobj 


    # ----------------------------------------------------------------------
    def get_quote(self, symbol: str) -> OEM_REPORT:
        doc="""
        获取指定代码的盘口行情.

        Args:
            symbol (str): 指定证券代码。可用的交易所代码如下：
                         * CFFEX: 中金所
                         * SHFE: 上期所
                         * DCE: 大商所
                         * CZCE: 郑商所
                         * INE: 能源交易所(原油)

        Returns: 
            :py:class:`~npsdk.Stockdrv.OEM_REPORT`: 返回一个盘口行情引用. 
            其内容将在 :py:meth:`~npsdk.api.NpSdkApi.wait_update` 时更新.

            注意: 在 npsdk 还没有收到行情数据包时, 此对象中各项内容为 None 或 0

        Example::

            # 获取 SH600000 的报价
            from nqsdk import NpSdkApi

            api = NpSdkApi()
            quote = api.get_quote("SH600000")
            print(quote.close)
            while api.wait_update():
                print(quote.close)

            # 预计的输出是这样的:
            0.0
            10.03
            10.03
            ...
        """

        """
        待开发功能： 检查参数 symbol的合法性，需要从服务器端下载代码表，检查symbol是否在代码表中
        """
        quote = OEM_REPORT() # 初始化结构体  
        quote.label = symbol   
        """
        for sybol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ].keys():
            # 如果最近一次读取实时行情成功，则取出最近一次获取的数据响应
            # 执行下面语句将更新策略程序quote变量引用的内存值，因为二者指向同一内存地址，用函数id(quote)可以查看实际内存地址
            quote = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][sybol][1]
        """
 
        # 向服务器端发送ws command命令请求, 命令格式:'代码=SH600000&类型=实时行情'
        npapiobj = NpApiDataObj()
        npapiobj.app_id = NPSDKAPI_APPID_NPSDKAPI
        npapiobj.message_id = NPSDKAPI_MESSAGEID_WSCOMMAND_REQ
        npapiobj.correlation_id = str(uuid.uuid4())
        npapiobj.reply_to = NPSDKAPI_MESSAGEID_WSCOMMAND_RES
        npapiobj.request_body = '代码=' + symbol + '&类型=实时数据'  # 类型不要写成'实时行情'
        self.put_command_queue(npapiobj) 

        # 向服务器端发送get_quote命令请求
        npapiobj = NpApiDataObj()
        npapiobj.app_id = NPSDKAPI_APPID_NPSDKAPI
        npapiobj.message_id = NPSDKAPI_MESSAGEID_GETQUOTE_REQ
        npapiobj.correlation_id = str(uuid.uuid4())
        npapiobj.reply_to = NPSDKAPI_MESSAGEID_GETQUOTE_RES
        npapiobj.request_body = symbol
        self.put_command_queue(npapiobj) #到数据池拿数据

        #print('npsdkapi:_response_queue.qsize():%s _command_queue.qsize():%s'%(self._response_queue.qsize(), self._command_queue.qsize()))
        
        time.sleep(NPSDKAPI_TIMEOUT_VAL)
        bret, npapiobjdes = self.get_response_queue() # 等数据池的数据消息
        if not bret: 
            npapiobjdes.response_type = NPSDKAPI_TIMEOUT

        bsuc = False
        if  npapiobjdes.response_type == NPSDKAPI_TIMEOUT:  # 向服务器发送命令后，等待命令响应超时。有可能超时后命令响应才到
            nowstr = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
            logger.info(" [.] err: got command(get_quote) response: %ssec timeout at: %s " %(NPSDKAPI_TIMEOUT_VAL, nowstr))
        
        elif npapiobjdes.app_id == NPSDKAPI_APPID_NPSDKDATA and \
            npapiobjdes.message_id == NPSDKAPI_MESSAGEID_GETQUOTE_RES  and \
            npapiobjdes.response_type == NPSDKAPI_SUCCESS:  # 向服务器发送命令后，收到命令响应（服务器端正常）。
            
            # 把字节流转成ctypes结构体
            quote.decode(npapiobjdes.response_body)
            if quote:
                bsuc = True
  
            nowstr = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
            logger.info(" [.] suc: got command(get_quote) response: %s(%s) close:%0.2f at: %s " %(quote.label, quote.name, quote.close, nowstr))

        elif npapiobjdes.app_id == NPSDKAPI_APPID_NPSDKDATA and \
             npapiobjdes.message_id == NPSDKAPI_MESSAGEID_GETQUOTE_RES  and \
             npapiobjdes.response_type == NPSDKAPI_ERROR:   # 向服务器发送命令后，收到命令响应（服务器端异常）。
            nowstr = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
            logger.info(" [.] err: got command(get_quote) response: %s at: %s " %(npapiobjdes.response_body, nowstr))
            
        #记录本次执行结果更新, ，以便在is_changing中做对比
        #列表第1项记录本次命令参数，第2项记录结果数据，第3项记录本次命令成功与否, 第4项纪录第一次get_quote(指定证券代码symbol)的结果df obj
        if not symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ].keys(): 
            self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol] = [[symbol],quote,bsuc,quote]
        else:
            lastobj = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][3] 
            self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol] = [[symbol],quote,bsuc,lastobj]

        # 不管是成功还是失败，每次均返回quote对象：成功返回实时行情，超时或者错误返回空quote
        return quote


    # ----------------------------------------------------------------------
    # 比较并且合并二次收到的K线数据DataFrame
    # 返回元组 (bool, pd.DataFrame)
    def _compare_merge_klines_dataframe(self, df1st, df2nd):
        try:
            if not isinstance(df1st, pd.DataFrame) or not isinstance(df2nd, pd.DataFrame):
                #logprtstr = 'invalide DataFrame Obj'
                #print(logprtstr); logger.info(logprtstr)             
                return False, None
            
            if len(df1st.index) == 0 and len(df2nd.index) == 0:
                #logprtstr = 'both dfs are empty DataFrame (df1st: %s )(df1st: %s )'%(len(df1st.index), len(df2nd.index))
                #print(logprtstr); logger.info(logprtstr)  
                return False, None

            """
            Test whether two objects contain the same elements.
            This function allows two Series or DataFrames to be compared against each other to 
            see if they have the same shape and elements. NaNs in the same location are considered equal.
            The row/column index do not need to have the same type, as long as the values are considered equal. 
            Corresponding columns must be of the same dtype.
            """
            if df1st.equals(df2nd): 
                #logprtstr = 'same DataFrame'
                #print(logprtstr); logger.info(logprtstr)  
                return False, None

            """
            # 1. df.empty ，这是DataFrame内置的属性，可以看到虽然调用简单，但他是最耗时的
            # 2. len(df)==0 ，这是通过Python内置len方法判断 DataFrame 的行数，相对来说速度比较快，是第1种的3倍
            # 3. len(df.index)==0 ，这是判断 DataFrame 的行索引的值数量，这已经到达纳秒级别了，是其中最快的方式当然，
            # 如果不是非常密集的调用，那么选哪种都无所谓。当你需要对程序进行性能调优时，就可以考虑选用上述的方式2或3。
  
            DateFrame.copy(deep=True) 
            data=DataFrame.copy(deep=False) 
            复制object的索引和数据

            当deep=True时(默认), 会创建一个新的对象进行拷贝. 修改这份拷贝不会对原有对象产生影响.
            当deep=False时, 新的对象只是原有对象的references. 任何对新对象的改变都会影响到原有对象
            """

            #合并二个不同的Klines DataFrame
            #print('@@@@@@@@@@@@@@@@@@@@@@@id df1st before merge: %s'%id(df1st))
            dfmerged = pd.merge(df1st, df2nd, how='right')   # on='time', how='right') 
            #dfmerged = df1st.append(df2nd)    
            #dfmerged.drop_duplicates(inplace = True)
            #print('@@@@@@@@@@@@@@@@@@@@@@@id df1st after merge: %s'%id(df1st))
            
            return True, dfmerged
        except Exception as e:
            logger.warning('_compare_merge_klines_dataframe() exception: %r'% e)
            return False, None

    # ----------------------------------------------------------------------
    # 把K线数据字节流转成DataFrame
    def _convert_klines_to_dataframe(self, symbol, kline_type, klines_rawdata)->pd.DataFrame:
        try:
            df_kline = None

            """
            if len(klines_rawdata) == 0: #如果输入0长度数据，为了在策略程序里第一次能访问df.iloc[-1],暂时造一行假数据
                klinebody = RCV_KLINE()
                klinebody.time = 0
                klinebody.open = 0
                klinebody.high = 0
                klinebody.low = 0
                klinebody.close = 0
                klinebody.volume = 0
                klinebody.amount = 0
                klinebody.temp = 0
                klines_rawdata = klinebody.encode()
            """

            kline_list_des = []
            count = int(len(klines_rawdata) / sizeof(RCV_KLINE)) # 计算多少根K线

            for i in range(count):
                klinebody = RCV_KLINE()
                klinebody.decode(klines_rawdata)                       #字节流 转成 ctypes结构体
                klines_rawdata = klines_rawdata[sizeof(RCV_KLINE) : ]  #删除一个结构体
                fieldval=[i, symbol, kline_type] # 放入K线序列号id，从0开始，指定证券代码, K线类型
                
                for field in self._fieldsOfKline:
                    if field == 'temp': #localize the timestamp to local datetime
                        val = time.localtime(klinebody.time)
                        val = time.strftime('%y-%m-%d %H:%M:%S', val)
                    else:
                        val = getattr(klinebody, field)
                    fieldval.append(val)
                    #fieldval.append(getattr(klinebody, field))
                kline_list_des.append(fieldval)

            #创建DataFrame:行索引默认从0开始，列索引里加入K线序列号id，也是从0开始, 指定证券代码, K线类型
            if len(kline_list_des) == 0: #创建空df
                df_kline = pd.DataFrame(columns=['id', 'label', 'ktype'] + self._fieldsOfKline)
            else:
                df_kline = pd.DataFrame(kline_list_des, columns=['id', 'label', 'ktype'] + self._fieldsOfKline)
            #df_kline.set_index(["time"], inplace=True)
            
        except Exception as e:
            logger.warning('_convert_klines_to_dataframe() exception: %r'% e)
        finally:
            return df_kline
    
    # ----------------------------------------------------------------------
    # 把K线数据字节流转成Klines结构体属性字段列表
    def _convert_klines_to_datalist(self, symbol, kline_type, klines_rawdata)->list:
        try:
            kline_list_des = []
            count = int(len(klines_rawdata) / sizeof(RCV_KLINE)) # 计算多少根K线

            for i in range(count):
                klinebody = RCV_KLINE()
                klinebody.decode(klines_rawdata)                      #字节流 转成 ctypes结构体
                klines_rawdata = klines_rawdata[sizeof(RCV_KLINE) : ]  #删除一个结构体
                fieldval=[i, symbol, kline_type] # 放入K线序列号id，从0开始，指定证券代码, K线类型
                
                for field in self._fieldsOfKline:
                    if field == 'time': #localize the timestamp to local datetime
                        val = time.localtime(klinebody.time)
                        val = time.strftime('%y-%m-%d %H:%M:%S', val)
                    else:
                        val = getattr(klinebody, field)
                    fieldval.append(val)
                    #fieldval.append(getattr(klinebody, field))
                kline_list_des.append(fieldval)

            #df_kline = pd.DataFrame(kline_list_des, columns=['id', 'label', 'ktype'] + self._fieldsOfKline)

        except Exception as e:
            logger.warning('_convert_klines_to_datalist() exception: %r'% e)
        finally:
            return kline_list_des

    # ----------------------------------------------------------------------
    def _compare_update_klines_datalist(self, kline_list_src, kline_list_des)-> list:
        # todo:
        return kline_list_des

    # ----------------------------------------------------------------------
    def get_kline_serial(self, symbol: str, kline_type: str = "1分钟线", data_length: int = 1000) -> pd.DataFrame:
        doc="""
        获取k线序列数据

        请求指定证券代码及周期的K线数据. 序列数据会随着时间推进自动更新

        Args:
            symbol (str): 指定证券代码
                * str: 一个证券代码
                * list of str: 证券代码列表 （一次提取多个证券的K线并根据相同的时间向第一个证券代码对齐) (暂时不支持）

            kline_type (str): K线类型字符串, 取值: 
                "1分钟线", "5分钟线", "15分钟线", "30分钟线", "60分钟线", "日线", "周线", "月线", "季线", "年线", "多日线"

            data_length (int): 需要获取的序列长度。默认200根, 返回的K线序列数据是从当前最新一根K线开始往回取data_length根。\
            每个序列最大支持请求 8964 ?(待测试确认) 个数据

         Returns::
            pandas.DataFrame: 本函数总是返回一个 pandas.DataFrame 实例. 行数=data_length, 包含以下列:

            * id:       (k线序列号，从0开始 )    * DataFrame里 新增字段
            * label:    (证券代码)              * DataFrame里 新增字段
            * ktype:    (K线类型)               * DataFrame里 新增字段

            * time:     (K线起点时间戳UTC，单位:秒数)   *** time字段是唯一业务索引 ***
            * open:     (K线起始时刻的最新价)
            * high:     (K线时间范围内的最高价)
            * low:      (K线时间范围内的最低价)
            * close:    (K线结束时刻的最新价)
            * volume:   (K线时间范围内的成交量)
            * amount:   (K线时间范围内的成交额)
            * temp:     (保留)

        Example1:

            # 获取 SH600000 的1分钟线
            from npsdkapi import NpSdkApi

            api = NpSdkApi()
            klines = api.get_kline_serial("SH600000", "1分钟线", 60)
            print(klines.iloc[-1].close)
            while True:
                api.wait_update()
                print(klines.iloc[-1].close)

            # 预计的输出是这样的:
            10.22
            10.22
            10.22
            ...
        """
        
        """
        待开发功能： 检查参数 symbol的合法性，需要从服务器端下载代码表，检查symbol是否在代码表中
        """
        try:  
            bsuc = False

            #是否正确的K线类型字符串格式
            if not kline_type in self._typesOfKline:
                raise Exception("K线类型 %s 错误, 请检查K线类型是否填写正确" % (kline_type))

            data_length = int(data_length)
            if data_length <= 0:
                raise Exception("K线数据序列长度 %d 错误, 请检查序列长度是否填写正确" % (data_length))
            if data_length > 8964:
                data_length = 8964
        
            """
            #创建一个空 klines DataFrame  
            """
            #创建空DataFrame:行索引默认从0开始，列索引里加入K线序列号id，也是从0开始, 指定证券代码, K线类型
            df_kline = self._convert_klines_to_dataframe(symbol, kline_type, b'') 

            # 向服务器端发送ws command命令请求, 命令格式:'代码=SH600000&类型=1分钟线&数量=3000'  申请3000根1分钟线
            npapiobj = NpApiDataObj()
            npapiobj.app_id = NPSDKAPI_APPID_NPSDKAPI
            npapiobj.message_id = NPSDKAPI_MESSAGEID_WSCOMMAND_REQ
            npapiobj.correlation_id = str(uuid.uuid4())
            npapiobj.reply_to = NPSDKAPI_MESSAGEID_WSCOMMAND_RES
            npapiobj.request_body = '代码=%s&类型=%s&数量=%s'%(symbol, kline_type, data_length)
            #npapiobj.request_body = '代码=' + symbol + '&类型=' + kline_type + '&数量=' + str(data_length)
            self.put_command_queue(npapiobj) 

            # 向服务器端发送get_kline_serial命令请求
            npapiobj = NpApiDataObj()
            npapiobj.app_id = NPSDKAPI_APPID_NPSDKAPI
            npapiobj.message_id = NPSDKAPI_MESSAGEID_GETKLINES_REQ
            npapiobj.correlation_id = str(uuid.uuid4())
            npapiobj.reply_to = NPSDKAPI_MESSAGEID_GETKLINES_RES
            npapiobj.request_body = symbol + '?' + kline_type  # 格式: 'SH600000?1分钟线'
            self.put_command_queue(npapiobj) #到数据池拿数据

            time.sleep(NPSDKAPI_TIMEOUT_VAL)
            bret, npapiobjdes = self.get_response_queue() # 等数据池的数据消息
            if not bret: 
                npapiobjdes.response_type = NPSDKAPI_TIMEOUT
                        
            if  npapiobjdes.response_type == NPSDKAPI_TIMEOUT:  # 向服务器发送命令后，等待命令响应超时。有可能超时后命令响应才到
                nowstr = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                logger.info(" [.] err: got command(get_kline_serial) response: %ssec timeout at: %s " %(NPSDKAPI_TIMEOUT_VAL, nowstr))

            elif npapiobjdes.app_id == NPSDKAPI_APPID_NPSDKDATA and \
                npapiobjdes.message_id == NPSDKAPI_MESSAGEID_GETKLINES_RES  and \
                npapiobjdes.response_type == NPSDKAPI_SUCCESS:  # 向服务器发送命令后，收到命令响应（服务器端正常）。
                
                #把K线数据原始字节流转成pandas.DataFrame
                #t = time.perf_counter()
                df_kline = self._convert_klines_to_dataframe(symbol, kline_type, npapiobjdes.response_body) 
                #print(f'_convert_klines_to_dataframe coast:{time.perf_counter() - t:.8f}s')
                if len(df_kline.index) != 0: # 判断df不为空, 收到数据并且转换成功才能把bsuc设为True
                    bsuc = True
                
                nowstr = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                logger.info(" [.] suc: got command(get_kline_serial) response: open:%0.2f high:%0.2f close:%0.2f at: %s " %(df_kline.iloc[-1]["open"], df_kline.iloc[-1]["high"], df_kline.iloc[-1]["close"], nowstr))

            elif npapiobjdes.app_id == NPSDKAPI_APPID_NPSDKDATA and \
                npapiobjdes.message_id == NPSDKAPI_MESSAGEID_GETKLINES_RES  and \
                npapiobjdes.response_type == NPSDKAPI_ERROR:   # 向服务器发送命令后，收到命令响应（服务器端异常）。
                nowstr = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                logger.info(" [.] err: got command(get_kline_serial) response: %s at: %s " %(npapiobjdes.response_body, nowstr))

            #记录本次执行结果更新, ，以便在is_changing中做对比
            #列表第1项记录本次命令参数，第2项记录结果数据，第3项记录本次命令成功与否, 第四项纪录第一次get_kline_serial(指定证券代码symbol)的结果df obj
            #以K线类型为key值 更新三级字典,以证券代码为key值 更新二级字典

            condition1 = symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ].keys()
            if condition1:
                condition2 = kline_type in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol].keys() 
           
            if condition1:
                if condition2:
                    #证券代码在二级字典并且K线类型在三级字典, 修改原纪录list
                    # 取出上一次df数据
                    lastobj = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][3]
                    if bsuc: # 如果成功收到数据
                        pass
                        """
                        #本次成功接收到K线数据，比较并且合并二次收到的K线数据 
                        #t = time.perf_counter()
                        bret, df_kline_merged = self._compare_merge_klines_dataframe(lastobj, df_kline)
                        #print(f'_compare_merge_klines_dataframe coast:{time.perf_counter() - t:.8f}s')
                        if bret: # 如果合并成功 则存放合并后的df
                            #清空DataFrame, 再按列赋值
                            lastobj.drop(lastobj.index,inplace=True) 
                            for col in df_kline_merged.columns:
                                lastobj[col] = df_kline_merged[col]
                            
                        """
                    else:
                        logger.info('前值存在，但是本次读取K线数据失败')

                    self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type] = \
                                            [[symbol, kline_type, data_length], df_kline, bsuc, lastobj] 
                else:
                    #证券代码在二级字典并且K线类型不在三级字典, 添加新纪录list，不影响别的K线类型list数据
                    #该K线类型第一次申请K线数据，收到数据无需比较, 所以保存结果后直接返回  
                    self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type] =\
                                            [[symbol, kline_type, data_length], df_kline, bsuc, df_kline]              

            else:
                #证券代码不在二级字典, 添加新纪录list  
                #该证券代码第一次申请K线数据，收到数据无需比较, 所以保存结果后直接返回                 
                klines_dict_bytype = {}
                klines_dict_bytype[kline_type] = [[symbol, kline_type, data_length], df_kline, bsuc, df_kline]
                self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol] = klines_dict_bytype

            # 不管是成功还是失败，每次均返回DataFrame对象：成功返回K线DataFrame，超时或者错误返回空DataFrame     
            return df_kline #如果接收数据失败，则会返回空df

        except Exception as e:
            logger.warning('get_kline_serial() exception: %r'% e)
        finally:
            return df_kline
             
    # ----------------------------------------------------------------------
    # 检查历史命令请求里是否有get_quote命令,有则执行
    def _run_quotes_tasks(self) -> None:
        if len(self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ]) > 0:
            
            for symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ].keys():
                cmdpara = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][0] #取出历史命令参数:列表
                self.get_quote(cmdpara[0]) # 找到历史命令，再次执行
                logger.info("执行历史命令：get_quote(%s)......: "% cmdpara[0])

        else: # 在历史命令记录里，没有找到一条get_quote命令，这种情况为 wait_update()函数前从未执行过get_quote命令
            pass


    # ----------------------------------------------------------------------
    # 检查历史命令请求里是否有get_klines命令,有则执行
    def _run_klines_tasks(self) -> None:
        if len(self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ]) > 0:

            for symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ].keys():
                for klinetypekey in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol].keys():
                    
                    cmdpara = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][klinetypekey][0] #取出历史命令参数:列表
                    #t = time.perf_counter()
                    self.get_kline_serial(cmdpara[0], cmdpara[1], cmdpara[2]) # 找到历史命令，再次执行
                    #print(f'get_kline_serial coast:{time.perf_counter() - t:.8f}s')

                    logger.info("执行历史命令：get_kline_serial(%s,%s,%s)......: "% (cmdpara[0], cmdpara[1], cmdpara[2]))

        else: # 在历史命令记录里，没有找到一条get_kline_serial命令，这种情况为 wait_update()函数前从未执行过get_kline_serial命令
            pass
    

    # ----------------------------------------------------------------------
    def wait_update(self) -> None:
        doc="""
        执行 调用本次wait_update()函数前的 历史命令，等待业务数据更新:
        * 再次执行历史命令 get_quote, get_klines, get_ticks...（暂时不包括 交易命令）
        * 后续用户可以调用 api.is_changing(参数) 查询用户关心的数据

        Args:
            None

        Returns:
            bool: 如果收到业务数据更新则返回 True, 如果到截止时间依然没有收到业务数据更新则返回 False

        """

        try:
            #print('等待数据更新......')

            #t = time.perf_counter()
                                    
            # 检查历史命令请求里是否有get_quote命令,有则执行
            self._run_quotes_tasks()

            # 检查历史命令请求里是否有get_klines命令,有则执行
            self._run_klines_tasks()

            time.sleep(0.1)

            #print(f'wait_update coast:{time.perf_counter() - t:.8f}s')

        except KeyboardInterrupt:
            print('Keyboard Interrupted')
            try:
                sys.exit(0)
            except SystemExit:
                os._exit(0)

    # ----------------------------------------------------------------------
    def is_changing(self, obj: Any, propkey: list = None) -> bool:
        doc="""
        判定obj最近是否有更新

        对比 获取的数据响应 和 前一次的数据响应，返回业务数据更新标志

        当业务数据更新导致 wait_update 返回后可以使用该函数判断 **本次业务数据更新是否包含特定obj或其中某个字段** 。

        关于判断K线更新的说明：
        当生成新K线时，其所有字段都算作有更新，若此时执行 api.is_changing(klines.iloc[-1]) 则一定返回True。
        Args:
        obj (any): 任意业务对象, 包括 get_quote 返回的 quote, get_kline_serial 返回的 DataFrame, get_account 返回的 account 等

        propkey (list of str): [必选]需要判断的字段，空列表表示全部属性字段
                                * （空列表）不指定: 当该obj下的任意字段有更新时返回True, 否则返回 False.
                                * list of str: 当该obj下的指定字段中的任何一个字段有更新时返回True, 否则返回 False.
        """

        try:
            #检查参数
            if obj is None:
                raise Exception("参数1为 None Obj, 请检查参数1是否填写正确")
                return False

            if not isinstance(propkey, list):
                propkey = [propkey] if propkey else []



                """
                实时行情: 数据结构体定义详见 Stockdrv.py
                """
            if  "Stockdrv.OEM_REPORT" in str(type(obj)): 
                # 要读取的属性字段 必须是正确的quote属性字段
                for pkey in propkey:
                    if not pkey in self._fieldsOfQuoteToCompared:
                        raise Exception("api.is_changing()函数参数2(%s)里有不正确的实时行情属性字段名"% propkey)
                        return False

                """
                for symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ].keys():
                    # 取出前一次获取的数据响应
                    current_quote = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][1]
                    lastobj = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][3]
                    bsuc = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][2]
                    print(lastobj.open, lastobj.high, lastobj.low, lastobj.close)
                    print(current_quote.open, current_quote.high, current_quote.low, current_quote.close)
                """

                for symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ].keys():
                    if  obj.label != symbol: 
                        continue

                    # 取出前一次获取的数据响应
                    current_quote = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][1]
                    lastobj = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][3]
                    bsuc = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][symbol][2]

                    #print('id(obj): %s id(lastobj): %s'%(id(obj),idoflastobj))
                    
                    # 如果前一次读取数据失败，再继续数据比较就没有意义，应该立即返回False
                    if not bsuc:
                        logger.info('前一次获取实时行情失败')
                        return False
                    # 如果id(obj)不匹配，再继续数据比较就没有意义，应该立即返回False
                    if id(obj) != id(lastobj):
                        logger.warning('实时行情对象内存地址已改变')
                        return False
                        
                    """
                    类变量引用中的属性字段比较： 用id(obj)可以获得obj的引用内存地址
                    策略程序里的类变量引用的内存值（执行wait_update前获得的）： quote = api.get_quote("SH600000")
                    和
                    最新获得的类变量引用的内存值（执行wait_update后获得的）：self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ]["SH600000"]
                    """
                    if len(propkey) == 0: #比较所有属性字段
                        fieldsToCompared = self._fieldsOfQuoteToCompared
                    else: #比较参数2里的属性字段
                        fieldsToCompared = propkey
                    is_changed = False
                    for pkey in fieldsToCompared:
                        if getattr(obj, pkey) != getattr(current_quote, pkey): # 如果有属性字段数据更新，则更新obj所有属性字段
                            logstrprt = "   sdk>>> 实时行情 数据更新(代码: %s 名称: %s) (属性字段: %s [旧:新] = %s : %s)"% (current_quote.label, current_quote.name, pkey, getattr(obj, pkey), getattr(current_quote, pkey))
                            logger.info(logstrprt)
                            if self._debug_npsdk: print(logstrprt); 
                            is_changed = True
                    if is_changed:
                        #memmove(addressof(obj), addressof(current_quote), sizeof(OEM_REPORT)) # 更新策略程序里的类变量引用的内存值
                        memmove(addressof(lastobj), addressof(current_quote), sizeof(OEM_REPORT)) # 更新策略程序里的类变量引用的内存值
                        return True 
                    else:
                        return False


            
                """
                DataFrame: DataFrame 是一个表格型的数据结构, 它可以被看做由Series组成的字典（共同用一个索引）
                Two-dimensional, size-mutable, potentially heterogeneous tabular data.
                1. K线数据
                2. ....
                """
            # Pandas Dataframe iloc[]->class 'pandas.core.series.Series'  
            # Pandas Dataframe->class 'pandas.core.frame.DataFrame'
            # DataFrame对象，仅比较最近二次取得的df是否一样，如果不一样，则检查最后一根K线的所有属性字段是否一样
            elif isinstance(obj, pd.DataFrame): 
                # 从DataFrame obj里取证券代码和K线类型
                try:
                    """
                    # K线Dataframe: 
                    """
                    if 'ktype' in obj.columns: 
                        for symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ].keys():
                            for kline_type in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol].keys():
                                
                                if id(obj) == id(self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][3]):
                                    
                                    #取出前一次读取K线成功后转成的DataFrame
                                    df_kline = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][1]
                                    lastobj = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][3]                            
                                    bsuc = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][2] #前一次读取K线是否成功

                                    if len(df_kline.index) == 0: #空DataFrame
                                        return False

                                    if not bsuc: #前一次从服务器读取读取K线失败，没必要继续比较
                                        return False

                                    # 如果id(obj)不匹配，再继续数据比较就没有意义，应该立即返回False
                                    #if id(obj) != id(lastobj):
                                    #    logger.info('is_changing()第一个参数不是数据更新对象或者被赋值过')
                                    #    return False
                                    
                                    #比较前后二个DataFrame是否完全一致
                                    if df_kline.equals(obj): 
                                        logger.info('前后二次收到的DataFrame一样')
                                        return False 
                                    else:
                                        """
                                        # # 以下情况判断K线发生变化 更新前一次收到的df
                                        # 1. 第一次没取到数据(空df) 
                                        # 2. 一根没走完的K线：在前后二次df中，该根K线的时间戳没变化但对应的属性字段发生变化
                                        # 3. 一根新K线：刚收到的df新增一根K线(时间戳变化了)
                                        """
                                        condition1 = (len(obj.index) == 0)
                                        condition2 = (not condition1) and (obj.iloc[-1]['time'] == df_kline.iloc[-1]['time'] and (not obj.iloc[-1].equals(df_kline.iloc[-1])))
                                        condition3 = (not condition1) and (obj.iloc[-1]['time'] != df_kline.iloc[-1]['time'])

                                        #print(condition1,condition2,condition3)

                                        if condition1 or condition2 or condition3:
                                            logstrprt = "   sdk>>> K线 数据更新(代码: %s K线类型: %s)"% (df_kline.iloc[-1].label, df_kline.iloc[-1].ktype)
                                            logger.info(logstrprt)
                                            if self._debug_npsdk: print(logstrprt); 

                                            #清空DataFrame, 再按列赋值
                                            lastobj.drop(lastobj.index,inplace=True) 
                                            for col in df_kline.columns:
                                                lastobj[col] = df_kline[col]
                                            #lastobj.set_index(["time"], inplace=True)

                                            """
                                            if condition2: 
                                               # 同根新值
                                                logstrprt = ">>>>>> K线 数据更新(代码: %s K线类型: %s) (***同根新值***)"% (df_kline.iloc[-1].label, df_kline.iloc[-1].ktype)
                                                print(logstrprt); logger.info(logstrprt)

                                            if condition3: 
                                                # 新增一根
                                                logstrprt = ">>>>>> K线 数据更新(代码: %s K线类型: %s) (***新增一根***)"% (df_kline.iloc[-1].label, df_kline.iloc[-1].ktype)
                                                print(logstrprt); logger.info(logstrprt)
                                            """

                                            return True
                                    break
                                else:
                                    pass
                        return False

                    else:
                        logger.warning('非K线数据的DataFrame对象')
                        return False

                except Exception as e:
                    logger.warning('isinstance(obj, pd.DataFrame) exception: %r'% e)
                    return False


                """
                #Series: Series 是一种类似于一维数组的对象.它由一组数据（各种Numpy数据类型）
                以及一组与之相关的数据标签（即索引）组成。
                One-dimensional ndarray with axis labels (including time series).
                1. K线数据
                2. ....
                """               
            elif isinstance(obj, pd.Series):
                # 从Series obj里取证券代码和K线类型
                #print('Series Object: id=%s\r\nname:\r\n%s\r\nindex:\r\n%s\r\nobj:\r\n%r'%(id(obj), obj.name, obj.index, obj))
                try:
                    """
                    # K线Series: 对应K线Series对象，is_changing函数参数1格式应该输入：df.iloc[-1]，
                    # 其它(df.iloc[0], df.iloc[-2]之类的输入)没有意义
                    """
                    if 'ktype' in obj.index: 
                        # 要读取的属性字段 必须是正确的kline属性字段
                        for pkey in propkey:
                            if not pkey in self._fieldsOfKline:
                                raise Exception("api.is_changing()函数参数2(%s)里有不正确的K线数据属性字段名"% propkey)
                                return False

                        #print('---is_changing---(%s : %s) ?'%(obj['label'], obj['ktype']))
                        """
                        len1 = len(self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ])
                        for symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ].keys():
                            len2 = len(self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol])
                            for kline_type in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol].keys():
                                df_kline = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][1]
                                lastobj = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][3]                            
                                bsuc = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][2] #前一次读取K线是否成功
                                print('len1:%s len2:%s symbol:%s ktype:%s bsuc:%r id(df_kline):%s id(lastobj):%s\r\n'%(len1, len2, symbol, kline_type, bsuc, id(df_kline),id(lastobj)))
                        """

                        for symbol in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ].keys():
                            if  obj['label'] != symbol: continue

                            for kline_type in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol].keys():
                                if obj['ktype'] != kline_type: continue

                                #取出前一次读取K线成功后转成的DataFrame
                                df_kline = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][1]
                                lastobj = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][3]                            
                                bsuc = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETKLINES_REQ][symbol][kline_type][2] #前一次读取K线是否成功

                                #print('obj%s %s %s'%(obj["label"], obj["ktype"], obj["time"]))
                                #print('lastobj%s %s %s'%(lastobj.iloc[-1]["label"], lastobj.iloc[-1]["ktype"], lastobj.iloc[-1]["time"]))
                                
                                if len(df_kline.index) == 0: #空DataFrame
                                    return False

                                if not bsuc: #前一次从服务器读取读取K线失败，没必要继续比较
                                    return False

                                if obj.values in lastobj.values: #检查Series是否在前一次DataFrame中
                                    # obj.name 是 该Series在DataFrame中的索引行号，用if obj.name in lastobj.index:也能判断
                                    
                                    # 属性字段time时间戳是业务唯一索引
                                    timestampe_src = obj['time']  #该obj的属性字段time时间戳
                                    if timestampe_src in df_kline['time'].values:
                                        index_src = lastobj['time'].tolist().index(timestampe_src)
                                        index_des = df_kline['time'].tolist().index(timestampe_src)

                                        #print('指定时间戳:(%s) Series Name:(%s) 源索引:(%s) 目的索引:(%s)'%(timestampe_src, obj.name, index_src, index_des))

                                        if len(propkey) == 0: #比较所有属性字段
                                            fieldsToCompared = self._fieldsOfKline
                                        else: #比较参数2里的属性字段
                                            fieldsToCompared = propkey

                                        is_changed = False
                                        for filed in fieldsToCompared:
                                            if obj[filed] != df_kline.iloc[index_des][filed]:
                                                logstrprt = "   sdk>>> K线指定行字段 数据更新(代码: %s K线类型: %s) (属性字段: %s [旧:新] =  %s : %s)"% (symbol, kline_type, filed, obj[filed], df_kline.iloc[index_des][filed])
                                                logger.info(logstrprt)
                                                if self._debug_npsdk: print(logstrprt); 
                                                is_changed = True

                                        """
                                        # 以下情况更新前一次收到的df
                                        # 1. 一根没走完的K线：在前后二次df中，该根K线的时间戳没变化但对应的属性字段发生变化
                                        # 2. 一根新K线：刚收到的df新增一根K线(时间戳变化了)
                                        """
                                        condition2 = is_changed
                                        condition3 = df_kline.iloc[-1]['time'] > df_kline.iloc[index_des]['time']

                                        if condition2 or condition3: 
                                             #清空DataFrame, 再按列赋值
                                            lastobj.drop(lastobj.index,inplace=True) 
                                            for col in df_kline.columns:
                                                lastobj[col] = df_kline[col]
                                            #lastobj.set_index(["time"], inplace=True) 

                                            """
                                            if condition2: 
                                                # 同根新值
                                                logstrprt = ">>>>>> K线指定行字段 数据更新(代码: %s K线类型: %s) (***同根新值***)"% (symbol, kline_type)
                                                print(logstrprt); logger.info(logstrprt)

                                            if condition3: 
                                                # 新增一根
                                                logstrprt = ">>>>>> K线指定行字段 数据更新(代码: %s K线类型: %s) (***新增一根***)"% (symbol, kline_type)
                                                print(logstrprt); logger.info(logstrprt)
                                            """
                                             
                                            return True 

                                    else: 
                                        logger.warning('该K线Series的时间戳不在刚收到的DataFrame中 %s  %s'%(obj['label'],obj['ktype']))
                                        #return False
                                        continue

                                else: 
                                    logger.warning('该K线Series对象不在前一次DataFrame中 %s  %s'%(obj['label'],obj['ktype']))
                                    continue
                                    #return False
                        
                        #print('---is_changing---(%s : %s): No'%(obj['label'], obj['ktype']))      
                        return False

                    else: #其它业务类型的Series obj
                        logger.warning('非K线数据的Series对象')
                        return False
                except Exception as e:
                    logger.warning('isinstance(obj, pd.Series) exception: %r'% e)
                    return False
                

            else: 
                print("不支持的(参数1)业务对象：%s"%type(obj))
                return False

            return False
        except Exception as e:  
            logger.warning('is_changing() exception: %r'% e)
            return False
        #except (KeyError, IndexError):
        #    return False


    """
    # ----------------------------------------------------------------------
    # 检查历史命令请求里是否有get_quote命令,有则执行并检查是否有数据更新
    def _is_quotes_changing(self) -> bool:
 
        if len(self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ]) > 0:
            
            print("历史实时行情命令：%s条"% len(self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ]))
            
            for key in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ].keys():
                # 逐条执行历史实时行情命令，并检查是否有更新

                # 取出最近一次获取的数据响应
                last_quote = self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ][key]
                
                current_quote = self.get_quote(key) # 找到历史命令，再次执行，获取最新数据

                print("执行历史命令：get_quote(%s)......: "% key)

                # 初始化 Quote是否有数据更新标志字典
                d = {}; 
                for field in self._fieldsOfQuoteToCompared:
                    d[field] = False
                self._quotes_items_changed[key] = d

                # 逐项比较二个Quote对应的属性字段值是否一样
                for field in self._fieldsOfQuoteToCompared:
                    if getattr(last_quote, field) != getattr(current_quote, field):
                       self._quotes_items_changed[key][field] = True
                       print("实时行情有更新...(%s): (%s) changed"% (current_quote.label, field))

                
            for key in self._npsdkapi_cmd_request_response[NPSDKAPI_MESSAGEID_GETQUOTE_REQ].keys():
                    for field in self._fieldsOfQuoteToCompared:
                        if self._quotes_items_changed[key][field]:
                            return True
 
            print("实时行情没有更新...(%s): "% current_quote.label)
                

        else: # 在历史命令记录里，没有找到一条get_quote命令，这种情况为 wait_update()函数前从未执行过get_quote命令
            return False

        return False
    """
        
            



from ctypes import *

#本示例尽量详细注释，没有注释的，请猜猜看
CHAR   = c_char
bool   = c_bool
BOOL   = c_int
INT8   = c_char
UCHAR  = c_ubyte
UINT8  = c_ubyte
BYTE   = c_ubyte
SHORT  = c_short
INT16  = c_short
USHORT = c_ushort
UINT16 = c_ushort
WORD   = c_ushort
INT    = c_int
INT32  = c_int
UINT   = c_uint
UINT32 = c_uint
LONG   = c_long
ULONG  = c_ulong
DWORD  = c_ulong
WCHAR  = c_wchar
UINT64 = c_ulonglong
FLOAT  = c_float
DOUBLE = c_double

LPCSTR  = POINTER(CHAR)
LPCVOID = POINTER(None)
NPSTR   = POINTER(CHAR)
LPSTR   = POINTER(CHAR)
PSTR    = POINTER(CHAR) 
NWPSTR  = POINTER(WCHAR)
LPWSTR  = POINTER(WCHAR)
PWSTR   = POINTER(WCHAR)

OEM_Ver2020  = 2020

class OEM_DATA_HEAD(Structure):     #通信头部 100 字节
    _pack_ = 1                      #对齐方式
    _fields_ = [                    #构体成员名称
        ("oemVer",    UINT),        #版本号
        ("type",      WCHAR * 10),  #各种子功能号：日线、5分钟线
        ("label",     WCHAR * 10),  #代码
        ("name",      WCHAR * 10),  #名称

        ("data",      UINT * 3),

        ("flag",      CHAR),        #-1设置某个值；0 服务器推送；1 读取本地数据
        ("askId",     INT),         #申请编号(内部功能)
        ("len",       INT),         #数据长度
        ("count",     INT),         #多少组数据
        ("temp1",     CHAR),        #预留
        ("temp2",     CHAR * 10)    #预留
        ]

    def __len__(self):
       return 100
       
    # ctypes结构体 转 字节流
    def encode(self): 
        return string_at(addressof(self), sizeof(self))

    # 字节流 转 ctypes结构体
    def decode(self, data):
        memmove(addressof(self), data, sizeof(self))
        return len(data)

class OEM_STKINFO(Structure):     #证券(100字节)
    _pack_ = 1
    _fields_ = [
    ('code',      WORD),            #序号
    ('market',    BYTE),            #市场MK_SH;MK_SZ，表明是上海、深圳等市场
    ('block',     BYTE),            #所属系统板块，BK_SHAG，BK_SZAG，表明是上海Ａ股，深圳Ａ股系统板块
    ('label',     WCHAR * 10),
    ('name',      WCHAR * 20),
    ('pointNum',  CHAR),            #小数点个数0、1、2、3
    ('hand',      WORD),            #每手股数
    ('last',      FLOAT),           #昨收
    ('limitUp',   FLOAT),           #涨停
    ('limitDown', FLOAT),           #跌停
    ('isIndex',   bool),            #是否指数*
    ('isDaPan',   bool),            #是否大盘*
    ('isStock',   bool),            #股票标识*
    ('bsNum',     CHAR),            #股票标识*
    ('pinYin',    CHAR * 12),
    ('temp',      CHAR * 5),        #
   ]

class OEM_MARKETINFO(Structure):       #市场内容(100字节)
    _pack_ = 1
    _fields_ = [
    ('mkId',       WORD),              #//市场代码, 'HS', 'SZ', 'JZ', 'HW' ..... 
    ('name',       WCHAR * 20),        #市场名称(英文简称，如 SHSE 表示上海交易所)
    ('tmCount',    SHORT),             #交易时段个数
    ('openTime',   SHORT * 8),         #开盘时间 1,2,3,4,5
    ('closeTime',  SHORT * 8),         #分钟数 比如 0X023A（570）表示 9:30)
    ('date',       UINT),              #数据日期（201301010）
    ('num',        WORD),              #该市场的证券个数
    ('temp',       CHAR * 18),
    ('stkInfo',    OEM_STKINFO * 10)
    ]

class OEM_REPORT(Structure):  #实时数据 400 字节，带*表示网际风内部数据，请跳过
    _pack_ = 1
    _fields_ = [
    ('label',      WCHAR * 10),              #代码
    ('name',       WCHAR * 10),              #名称
    ('time',       UINT),                    #成交时间(UTC时间，可以百度搜索相关转换)
    ('foot',   INT),                         #最后处理分钟数
    ('openDate',   UINT),                    #市场开盘日期零点
    ('openTime',   UINT),                    #市场开盘时间
    ('closeDate',  UINT),                    #市场日线日期零点

    ('open',       FLOAT),                   #今日开盘
    ('high',       FLOAT),                   #今日最高
    ('low',        FLOAT),                   #今日最低
    ('close',      FLOAT),                   #最新价格
    ('volume',     FLOAT),                   #总成交量
    ('amount',     FLOAT),                   #总成交金额
    ('inVol',      FLOAT),                   #外盘*
    ('pricesell',  FLOAT * 10),              #申卖价1,2,3,4,5
    ('volsell',    FLOAT * 10),              #申卖量1,2,3,4,5      
    ('vsellCha',   FLOAT * 10),              #申卖量变化*
    ('pricebuy',   FLOAT * 10),              #申买价1,2,3,4,5      
    ('volbuy',     FLOAT * 10),              #申买量1,2,3,4,5       
    ('vbuyCha',    FLOAT * 10),              #申买量变化*
    ('jingJia',    SHORT),                   #集合竞价
    ('avPrice',    FLOAT),                   #期货的结算价（平均价）
    ('isBuy',      CHAR),                    #0 下跌；1 上涨或平盘*
    ('nowv',       FLOAT),                   #现量
    ('nowa',       FLOAT),                   #分笔成交额(期货仓差)

    ('change',     FLOAT),                  #换手率*
    ('weiBi',      FLOAT),                  #委比*
    ('liangBi',    FLOAT),                  #量比*
    ('temp',       CHAR * 45),
    ]

    def __len__(self):
       return 400

    # ctypes结构体 转 字节流
    def encode(self): 
        return string_at(addressof(self), sizeof(self))

    # 字节流 转 ctypes结构体
    def decode(self, data):
        memmove(addressof(self), data, sizeof(self))
        return len(data)

class OEM_TRACE(Structure):  #分笔 100 字节
    _pack_ = 1
    _fields_ = [
    ('time',      UINT),
    ('close',     FLOAT),      #现价
    ('volume',    FLOAT),      #成交量
    ('amount',    FLOAT),      #成交额
    ('traceNum',  INT),
    ('pricebuy',  FLOAT * 5),  #申买价1,2,3,4,5
    ('volbuy',    FLOAT * 5),
    ('pricesell', FLOAT * 5),
    ('volsell',   FLOAT * 5),
    ]

    def __len__(self):
       return 108

class RCV_KLINE(Structure):
    _pack_ = 1
    _fields_ = [
    ('time',     UINT),           #时间
    ('open',     FLOAT),          #开盘
    ('high',     FLOAT),          #最高
    ('low',      FLOAT),          #最低
    ('close',    FLOAT),          #现价
    ('volume',   FLOAT),          #成交量
    ('amount',   FLOAT),          #成交额
    ('temp',     FLOAT),          #预留
    ]

    # ctypes结构体 转 字节流
    def encode(self): 
        return string_at(addressof(self), sizeof(self))

    # 字节流 转 ctypes结构体
    def decode(self, data):
        memmove(addressof(self), data, sizeof(self))
        return len(data)

class OEM_SPLIT_HEAD(Structure):      #除权头  100 字节
    _pack_ = 1
    _fields_ = [
    ("label",      WCHAR * 10),   #代码
    ("name",       WCHAR * 10),   #名称
    ("num",        WORD),         #数量
    ("volume",     UINT),         #系统内部
    ("temp",       CHAR * 54),    #预留
    ]

class OEM_SPLIT(Structure):           #除权 100 字节
    _pack_ = 1                       
    _fields_ = [                     
    ('time',       UINT),             #时间
    ("give",       FLOAT),            #每股送
    ("allocate",   FLOAT),            #每股配
    ("price",      FLOAT),            #每股配价
    ("earnings",   FLOAT),            #每股红利
    ("explain",    WCHAR * 40),       #描述文本(预留)
   ]

class RCV_FINANCE(Structure):       #财务 250 字节
    _pack_ = 1
    _fields_ = [
    ("label",    WCHAR * 10),         #代码
    ("name",     WCHAR * 10),         #名称
    ('time',     INT),                #发布日期		年月日
    ('baoGao',   INT),                #报告日期
    ('shangShi', INT),                #上市日期
    ('meiGuShouYi', FLOAT),           #每股收益(元)，净利润/总股本
    ('shuiHou',  FLOAT),              #每股净资产(元)
    ('jzcsyl',   FLOAT),              #净资产收益率(%)
    ('mgjyxj',   FLOAT),              #每股经营现金
    ('mggjj',    FLOAT),              #每股公积金(元)
    ('mgwfp',    FLOAT),              #每股未分配利润(元)
    ('gdqybl',   FLOAT),              #股东权益比率(%)
    ('jlrtb',    FLOAT),              #净利润同比		10
    ('zysytb',   FLOAT),              #主营收益同比
    ('xsmll',    FLOAT),              #销售毛利率
    ('meiGuJingZhi', FLOAT),          #调整每股净资(元)
    ('zongZC',   FLOAT),              #总资产(万元)
    ('ldzc',     FLOAT),              #流动资产(万元)
    ('guDingZC', FLOAT),              #固定资产(万元)
    ('wuXingZC', FLOAT),              #无形资产(万元)
    ('ldfz',     FLOAT),              #流动负债(万元)
    ('cqfz',     FLOAT),              #长期负债(万元)
    ('zfz',      FLOAT),              #总负债	20
    ('quanYi',   FLOAT),              #股东权益(净资产 万元)
    ('ziBenGongJi', FLOAT),           #资本公积金(万元)
    ('xianJin',   FLOAT),             #经营现金流量
    ('tzxjl',     FLOAT),             #投资现金流量
    ('czxjl',     FLOAT),             #筹资现金流量
    ('xjzje',     FLOAT),             #现金增加额
    ('shouRu',    FLOAT),             #主营业务收入(万元)
    ('shouRuZhu', FLOAT),             #主营利润(万元)
    ('liRun',     FLOAT),             #营业利润(万元)
    ('shouYi',    FLOAT),             #投资收益		30
    ('qtlr',      FLOAT),             #营业外收支(万元)
    ('zongLiRun', FLOAT),             #利润总额(万元)
    ('jingLiRun', FLOAT),             #净利润(千元)
    ('weiFenPei', FLOAT),             #未分配利润(万元)
    ('zongGu',    FLOAT),             #总股本(万股)
    ('wxsAg',     FLOAT),             #无限售股合计
    ('liuTongAG', FLOAT),             #流通A股(万股)
    ('bGu',       FLOAT),             #B股(万股)
    ('hGu',       FLOAT),             #境外上市股(万股)
    ('qtltg',     FLOAT),             #其它流通股	40
    ('xsghj',     FLOAT),             #限售股合计
    ('gjg',       FLOAT),             #国家股(万股)
    ('frcg',      FLOAT),             #国有法人股
    ('frg',       FLOAT),             #境内法人股(万股)
    ('jnzrrg',    FLOAT),             #境内自然人股
    ('fqrfrg',    FLOAT),             #其他发起人股(万股)
    ('mjfrg',     FLOAT),             #募集法人股
    ('jwfrg',     FLOAT),             #境外法人股
    ('jwzrrg',    FLOAT),             #境外自然人股
    ('yxg',       FLOAT),             #优先股或其他	50(大智慧204字节结构)
    ('temp',      CHAR * 6),
]

class OEM_F10INFO(Structure):         #F10资料 50 字节
    _pack_ = 1
    _fields_ = [
    ('title',  WCHAR * 12),           #信息标题         
    ('from',   INT),                  #文件位置         
    ('len',    INT),                  #长度        
    ('temp',   CHAR * 18),            #预留         
]

class CText(Structure):               #转换文本
    _pack_ = 1
    _fields_ = [
    ('data',   WCHAR * 10240000),
    ]



# from distutils.core import setup
from setuptools import setup, find_packages
import npsdkk

with open("README.md", mode="r", encoding='utf-8') as fh:
    long_description = fh.read()
    
setup(
    name='npsdkk',
    description="Nezip Python SDK",
    version='0.9.2',
    author='xxg',
    author_email='auiauk@qq.com', #'752299578@qq.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    # packages=setuptools.find_packages(include=['npsdkk', 'npsdkapi.py']),
    packages=find_packages(),
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Communications', 'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking'
    ],
    zip_safe=True,
    python_requires='>=3.6',
    
)


#:!/usr/bin/env python
#:  -*- coding: utf-8 -*-
__author__ = 'xxg'

from Stockdrv import *
from ctypes import *

"""
NPSDK 全局常量定义
"""

NPSDKAPI_APPID_NPSDKAPI  = 'npsdkapi'
NPSDKAPI_APPID_NPSDKDATA = 'npsdkdata'

NPSDKAPI_MESSAGEID_GETQUOTE_REQ     = 'get_quote'
NPSDKAPI_MESSAGEID_GETQUOTE_RES     = 'get_quote_response'
NPSDKAPI_MESSAGEID_GETKLINES_REQ    = 'get_klines'
NPSDKAPI_MESSAGEID_GETKLINES_RES    = 'get_klines_response'

NPSDKAPI_MESSAGEID_WSCOMMAND_REQ    = 'ws_command_request'
NPSDKAPI_MESSAGEID_WSCOMMAND_RES    = 'ws_command_response'    

NPSDKAPI_SUCCESS        = 'npsdkapi_success'
NPSDKAPI_ERROR          = 'npsdkapi_error'
NPSDKAPI_TIMEOUT        = 'npsdkapi_timeout'

NPSDKAPI_TIMEOUT_VAL    = 0.5 # 单位:secs, 命令请求超时阈值

NPSDKAPI_WS_CONNECTED           = 10
NPSDKAPI_WS_DISCONNECTED        = 0

NPSDKAPI_QUEUESZIE_WARNING      = 5000 # Websocket接收数据队列大小监控阈值，队列长度太大表示消息已经堆积

NPSDKAPI_PLAYMODE_WEBSOCKET     = 'websocket'   
NPSDKAPI_PLAYMODE_BACKTEST     = 'backtest'


#from tqsdk.diff import _get_obj
#from entity import Entity

""" Pika BasicProperties  详细说明https://www.pianshen.com/article/19451397954/
        content_type            用于描述消息内容的数据格式，如：text/plain
        content_encoding        消息内容编码
        headers                 设置消息的header,类型为Map<String,Object>
        delivery_mode           1（nopersistent）非持久化，2（persistent）持久化
        priority                消息的优先级
        correlation_id          关联ID
        reply_to                用于指定回复的队列的名称
        expiration              消息的失效时间
        message_id              消息ID
        timestamp               消息的时间戳
        type                    类型
        user_id                 用户ID
        app_id                  应用程序ID
        cluster_id              集群ID

"""

class AplEventContentObject(object):
    """ 
    NpSdkApi user-defined event content object class (application level)
    """
    def __init__(self):
        self.apl_tid_source = ''        # application level ID,such as 'DATA_POOL_ID','WS_SOCKET_ID','NPSDKAPI_ID'
        self.apl_tid_destination = ''
        self.apl_event_name = None
        self.apl_event_body = ''


class NpApiDataObj(object):
    """ 
    NpApi user-defined data object class 
    """
    def __init__(self):
        
        """
        应用ID, 用来标识应用 
                * NPSDKAPI_APPID_NPSDKAPI   策略程序
                * NPSDKAPI_APPID_NPSDKDATA  数据服务器
                * ......
        """
        self.app_id = None

        """
        消息ID，用来描述函数命令： 
                * NPSDKAPI_MESSAGEID_GETQUOTE_REQ            获取实时数据命令
                * NPSDKAPI_MESSAGEID_GETQUOTE_RES            获取实时数据响应
                * NPSDKAPI_MESSAGEID_GETKLINES_REQ           获取K线数据命令
                * NPSDKAPI_MESSAGEID_GETKLINES_RES           获取K线数据响应
                *
                * NPSDKAPI_MESSAGEID_WSCOMMAND_REQ           ws command 请求命令
                * NPSDKAPI_MESSAGEID_WSCOMMAND_RES           ws command 请求命令响应
                * .....
        """
        self.message_id = None 

        # 回复队列
        self.reply_to = None

        #: 关联ID
        self.correlation_id = None

        #: 请求消息参数体
        self.request_body = None

        """
        响应消息类型, 对应BasicProperties里的tpye
                * NPSDKAPI_SUCCESS:     命令请求成功
                * NPSDKAPI_ERROR:       命令请求出错
                * NPSDKAPI_TIMEOUT:     命令请求超时
                * ....
        """
        self.response_type = None 

        #: 响应消息体
        self.response_body = None

    def _instance_entity(self, path):
        pass
        #super(Quote, self)._instance_entity(path)
        #self.trading_time = copy.copy(self.trading_time)
        #self.trading_time._instance_entity(path + ["trading_time"])

"""
class Quote(Entity):
    #Quote 是一个行情对象 

    def __init__(self, api):
        self._api = api
        #: 行情从交易所发出的时间(北京时间), 格式为 "2017-07-26 23:04:21.000001"
        self.datetime = ""
        #: 卖一价
        self.ask_price1 = float("nan")
        #: 卖一量
        self.ask_volume1 = 0
        #: 买一价
        self.bid_price1 = float("nan")
        #: 买一量
        self.bid_volume1 = 0
        #: 卖二价
        self.ask_price2 = float("nan")
        #: 卖二量
        self.ask_volume2 = 0
        #: 买二价
        self.bid_price2 = float("nan")
        #: 买二量
        self.bid_volume2 = 0
        #: 卖三价
        self.ask_price3 = float("nan")
        #: 卖三量
        self.ask_volume3 = 0
        #: 买三价
        self.bid_price3 = float("nan")
        #: 买三量
        self.bid_volume3 = 0
        #: 卖四价
        self.ask_price4 = float("nan")
        #: 卖四量
        self.ask_volume4 = 0
        #: 买四价
        self.bid_price4 = float("nan")
        #: 买四量
        self.bid_volume4 = 0
        #: 卖五价
        self.ask_price5 = float("nan")
        #: 卖五量
        self.ask_volume5 = 0
        #: 买五价
        self.bid_price5 = float("nan")
        #: 买五量
        self.bid_volume5 = 0
        #: 最新价
        self.last_price = float("nan")
        #: 当日最高价
        self.highest = float("nan")
        #: 当日最低价
        self.lowest = float("nan")
        #: 开盘价
        self.open = float("nan")
        #: 收盘价
        self.close = float("nan")
        #: 当日均价
        self.average = float("nan")
        #: 成交量
        self.volume = 0
        #: 成交额
        self.amount = float("nan")
        #: 持仓量
        self.open_interest = 0
        #: 结算价
        self.settlement = float("nan")
        #: 涨停价
        self.upper_limit = float("nan")
        #: 跌停价
        self.lower_limit = float("nan")
        #: 昨持仓量
        self.pre_open_interest = 0
        #: 昨结算价
        self.pre_settlement = float("nan")
        #: 昨收盘价
        self.pre_close = float("nan")
        #: 合约价格变动单位
        self.price_tick = float("nan")
        #: 合约价格小数位数
        self.price_decs = 0
        #: 合约乘数
        self.volume_multiple = 0
        #: 最大限价单手数
        self.max_limit_order_volume = 0
        #: 最大市价单手数
        self.max_market_order_volume = 0
        #: 最小限价单手数
        self.min_limit_order_volume = 0
        #: 最小市价单手数
        self.min_market_order_volume = 0
        #: 标的合约
        self.underlying_symbol = ""
        #: 行权价
        self.strike_price = float("nan")
        #: 合约类型
        self.ins_class = ""
        #: 交易所内的合约代码
        self.instrument_id = ""
        #: 交易所代码
        self.exchange_id = ""
        #: 合约是否已下市
        self.expired = False
        #: 交易时间段
        self.trading_time = TradingTime(self._api)
        #: 到期具体日
        self.expire_datetime = float("nan")
        #: 到期月
        self.delivery_month = 0
        #: 到期年
        self.delivery_year = 0
        #: 期权方向
        self.option_class = ""
        #: 品种代码
        self.product_id = ""
        #: ETF实时单位基金净值
        self.iopv = float("nan")

    def _instance_entity(self, path):
        super(Quote, self)._instance_entity(path)
        self.trading_time = copy.copy(self.trading_time)
        self.trading_time._instance_entity(path + ["trading_time"])
"""

